/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.data;

import java.util.*;

import javax.vecmath.Point3f;
import javax.vecmath.Quat4f;
import javax.vecmath.Vector3f;

import netacka3d.misc.Logger;

/**
 * Pojemnik na dane gry. Przede wszystkim wstążki.
 * Jeden obiekt obsługuję jedną rundę.
 */
public final class ArenaData {
    // TODO private final arena - załadowana plansza (geometria planszy)

    private final MatchSettings matchSettings;

    private final TreeMap<Integer, ArrayList<Segment>> ribbons =
            new TreeMap<Integer, ArrayList<Segment>>();
    
    /** czy sprawdzamy kolizje */
    private final boolean checkCollisions;
    
    private final Octree octree;

    /**
     * Konstruktor.
     */
    public ArenaData(MatchSettings matchSettings, boolean checkCollisions) {
        this.matchSettings = matchSettings.getCopy(); 
        // nie kopiujemy areny!
        assert this.matchSettings.arena == matchSettings.arena;
        this.checkCollisions = checkCollisions;
        if (checkCollisions) {
            this.octree = new Octree(matchSettings);
        } else {
            this.octree = null;
        }
    }

    /**
     * Oblicza t-ty segment wstęgi gracza o podanym id, 
     * korzystając z segmentu t-1 i podanego kierunku skrętu.
     * 
     * TODO okomentować rachunki!
     * 
     * @param id numer gracza
     * @param t (numer segmentu do obliczenia)
     * @param dir kierunek skrętu
     * @param valid czy dane są zatwierdzone przez serwer
     * @return obliczony segment
     */
    public Segment calcNewSegment(int id, int t, Direction dir,
            boolean valid) {
        if (t == 0)
            throw new UnsupportedOperationException();
        Segment prevSeg = getSegment(id, t - 1);
        return prevSeg.calcNext(dir, valid, matchSettings);
    }
    
    /**
     * Sprawdza, czy dany segment koliduje z czymś, a jeśli tak,
     * to w którym miejscu.
     * 
     * @param seg segment, dla którego liczymy kolizje
     * @param playerId id gracza - właściciela wstążki (potrzebny, żeby 
     *                 pominąć jego ostatni segment)
     * @return czas zderzenia względem segmentu (przedział [0,1]; -1 - brak
     * zderzenia)
     */
    public float checkCollision(Segment seg, int playerId) {
        assert checkCollisions; 
        
        /*for (Vector3f v: seg.vertices) {
            if (v.x < 0 || v.x > matchSettings.arenaSize ||
                v.y < 0 || v.y > matchSettings.arenaSize ||
                v.z < 0 || v.z > matchSettings.arenaSize) 
            {
                return 1;
            }
        }*/
        
        if (!matchSettings.collide)
            return -1;
                
        ArrayList<Segment> ribbon = ribbons.get(playerId);
        Segment lastSegment = (ribbon.size()==0) ?
                              null : ribbon.get(ribbon.size()-1);
        if (octree.checkSegment(seg, lastSegment))
            return 1;
        
        return -1;
    }

    /**
     * Ustawia t-ty segment wstęgi.
     * 
     * @param player gracz kontrolujący wstęgę
     * @param t czas (numer segmentu)
     * @param seg segment do ustawienia
     */
    public void setSegment(int id, int t, Segment seg) {
        ArrayList<Segment> ribbon = ribbons.get(id);
        while (ribbon.size() < t + 1)
            ribbon.add(null);
        
        if (checkCollisions && seg != null) {
            // jeśli zamieniamy segment, to nie sprawdzamy kolizji
            //assert ribbon.get(t) == null; //to nieprawda przy resendach! 
            octree.insertSegment(seg);
        }
        ribbon.set(t, seg);
    }

    /**
     * Zwraca t-ty segment wstęgi gracza o podanym id (lub null, jeśli 
     * takiego nie mamy).
     * 
     * @param id numer gracza
     * @param t czas (numer segmentu)
     * @return żądany segment (lub null, jeśli go nie ma)
     */
    public Segment getSegment(int id, int t) {
        ArrayList<Segment> ribbon = ribbons.get(id);
        if (ribbon != null && ribbon.size() >= t + 1)
            // segmenty są niemutowalne, więc nie trzeba ich kopiować
            return ribbon.get(t);
        else
            return null;
    }
    
    /** Wylosuj segmenty początkowe, tak, żeby nie były za blisko siebie. */
    public void putFirstSegments() {
        
        long time = System.currentTimeMillis(); 
        float[] startingPs = matchSettings.arena.startingPoints;
        HashSet<Vector3f> positions = new HashSet<Vector3f>();
        
        for (int pId : ribbons.keySet()) {
            //losowanie pozycji startowej
            Segment seg = null;
            while (true) {
                Vector3f pos;
                if (startingPs.length == 0) {
                    // losowanie pozycji wewnątrz bounding-boxa planszy (cube)
                    Vector3f delta = matchSettings.arena.getDelta();
                    pos = new Vector3f((float)Math.random()*delta.x,
                                       (float)Math.random()*delta.y,
                                       (float)Math.random()*delta.z);
                    pos.scaleAdd(8, delta); // pos = pos*0.8 + delta
                    pos.scale((float) 0.1);
                    pos.add(matchSettings.arena.getOrigin());
                }
                else {
                    // losowanie z ustalonych
                    int numPositions = startingPs.length / 3;
                    int index = (int) Math.floor(Math.random() * numPositions);
                    Logger.d("wylosowałem pozycję " + index + " z " + numPositions);
                    assert index <= numPositions;
                    pos = new Vector3f(startingPs[index*3],
                                       startingPs[index*3 + 1], 
                                       startingPs[index*3 + 2]);
                }
                pos.scale(matchSettings.arenaSize);
                
                Quat4f heading = new Quat4f(
                        (float)Math.random(), 
                        (float)Math.random(), 
                        (float)Math.random(),
                        (float)Math.random());
                heading.normalize();
                
                boolean good = true;
                for (Vector3f pos2 : positions) {
                    if (pos.epsilonEquals(pos2, matchSettings.arenaSize/10))
                        good = false;
                }
                if (!good) 
                    continue;
                
                if (checkCollisions) {
                    good = false;
                    for (int i=0; i<10; ++i) {
                        matchSettings.segmentLength *= 10;
                        Segment longSeg = new Segment(pos, heading,
                                    Direction.STRAIGHT, true, matchSettings);
                        matchSettings.segmentLength /= 10;
                        if (checkCollision(longSeg, pId) < 0) {
                            good = true; //jeśli nie ma kolizji, to seg jest ok
                            break;
                        }
                    }
                }
                if (!good) 
                    continue;

                positions.add(pos);
                seg = new Segment(pos, heading, Direction.STRAIGHT,
                        true, matchSettings);
                
                break;
            }            
            
            setSegment(pId, 0, seg);
        }
        
        Logger.i("Losowanie pozycji trwalo "+(System.currentTimeMillis()-time)+"ms");
    }
    
    /** Dodaj gracza o podanym id. */
    public void addPlayer(int id) {
        assert (!ribbons.containsKey(id));
        ribbons.put(id, new ArrayList<Segment>());
    }

    public boolean hasPlayer(int id) {
        return ribbons.containsKey(id);
    }
    
    public Integer[] getPlayerIds() {
        return ribbons.keySet().toArray(new Integer[0]);
    }
    
    public int ribbonCount() {
        return ribbons.size();
    }
    
    /** Zwraca tablicę segmentów dla danego kroku t, w takiej kolejności jak
     * aktywni gracze (jeśli danego segmentu nie ma, w tablicy jest null).
     */
    public Segment[] getSegments(int t) {
        Segment[] result = new Segment[ribbons.size()];
        int i = 0;
        for (ArrayList<Segment> ribbon: ribbons.values())
            result[i++] = t < ribbon.size() ? ribbon.get(t) : null;
        assert i == ribbons.size(); 
        return result;
    }
    
    /** zwraca mapę: playerId->segment dla chwili t.
     * (jeśli segmentu nie ma, to w mapie jest null) */
    public TreeMap<Integer, Segment> getSegmentsAsMap(int t) {
        TreeMap<Integer, Segment> result = new TreeMap<Integer, Segment>();
        
        for (Map.Entry<Integer, ArrayList<Segment>> entry: 
            ribbons.entrySet())
        {
            ArrayList<Segment> ribbon = entry.getValue();
            result.put(entry.getKey(), 
                    t < ribbon.size() ? ribbon.get(t) : null);
        }
        return result;
    }
    
    public String toString() {
        StringBuilder b = new StringBuilder();
        b.append("ArenaData:\n");
        for (Map.Entry<Integer, ArrayList<Segment>> entry: 
            ribbons.entrySet())
        {
            b.append(" Player "+entry.getKey().toString()+": \n");
            for (int i = 0; i < entry.getValue().size(); ++i)
            {
                b.append(String.format("  %d: %s\n", i, entry.getValue().get(i)));
            }
        }
        return b.toString();
    }
}
