/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.data;

import javax.vecmath.Vector3f;

/**
 * Funkcje do liczenia przecięcia trójkątów i odcinków
 * @author Pawel
 *
 */
public class IntersectionDetector {
    static float EPSILON = 0.005f; 
    
    /**
     * Sprawdza czy trójkąt odcinek w R^3 się przecinają.
     * @param q1 pierwszy wierzchołek trójkąta
     * @param q2 drugi ,,
     * @param q3 trzeci ,,
     * @param p1 jeden koniec odcinka
     * @param p2 drugi ,,
     * @return prawda, jeśli sie przecinają
     */
    public static boolean intersectTriangleLine(Vector3f q1, Vector3f q2,
            Vector3f q3, Vector3f p1,Vector3f p2) 
    {
        
        Vector3f n = Util3d.crossDelta(q2, q1, q3, q1); 
        Vector3f p1p2 = new Vector3f(p1);
        p1p2.sub(p2);
        Vector3f p1q1 = new Vector3f(p1);
        p1q1.sub(q1);
        float d1 = p1p2.dot(n);
        float d2 = p1q1.dot(n);
        if (Math.abs(d1) < EPSILON) {
            // odcinek równoległy do płaszczyzny
            if (Math.abs(d2) > EPSILON)
                return false;
            else {
                // odcinek na płaszczyźnie 
                return intersectLines(p1, p2, q1, q2) ||
                       intersectLines(p1, p2, q2, q3) ||
                       intersectLines(p1, p2, q3, q1);
            }
        }
        
        // t - miejsce przecięcia między p1 a p2
        // (jeśli 0<t<1, to jesteśmy na odcinku)
        float t = d2/d1;
        
        if(t<0.0-EPSILON || t>1.0+EPSILON)
            return false;

        // p - punkt przecięcia prostej z płaszczyzną trójkąta.
        Vector3f p = new Vector3f(p1);
        p1p2.scale(t);
        p.sub(p1p2);        

        // teraz patrzymy, czy p jest wewnątrz trójkąta
        
        Vector3f c1 = Util3d.crossDelta(q2, q1, p, q1),
                 c2 = Util3d.crossDelta(q3, q2, p, q2),
                 c3 = Util3d.crossDelta(q1, q3, p, q3);

        float dot1 = c1.dot(c2),
              dot2 = c2.dot(c3),
              dot3 = c3.dot(c1);

        if(dot1 >= 0 && dot2 >=0 && dot3 >= 0)
            return true;

        return false;
    }
    
    /** Sprawdza, czy przecinają się odcinki p1p2 i q1q2
     */
    private static boolean intersectLines(Vector3f p1, Vector3f p2,
            Vector3f q1, Vector3f q2)
    {
        Vector3f c1 = Util3d.crossDelta(p2, p1, q1, p1),
                 c2 = Util3d.crossDelta(p2, p1, q2, p1),
                 c3 = Util3d.crossDelta(q2, q1, p1, q1),
                 c4 = Util3d.crossDelta(q2, q1, p2, q1);
        
        float dot1 = c1.dot(c2),
              dot2 = c3.dot(c4);
        if(dot1 < -EPSILON && dot2 < -EPSILON)
            return true;
        return false;
    }            
}
