/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.data;

import java.util.*;

import javax.vecmath.Vector3f;

import netacka3d.data.arena.Mesh;
import netacka3d.data.arena.MeshData;
import netacka3d.data.arena.Mesh.Tri;
import netacka3d.data.arena.Mesh.Vertex;

/**
 * Drzewo ósemkowe przechowujące segmenty na planszy. Pozwala na wyciągnięcie
 * dla danego segmentu tych segmentów, które są w miarę blisko niego, co 
 * powinno przyspieszać liczenie kolizji.
 * 
 * @author humpolec
 */
public final class Octree {
    private static final int MAX_TRIANGLES_PER_NODE = 50;
    private float min_node_width;
    
    private Node root;
    
    public Octree(MatchSettings mS) {
        Vector3f origin = mS.arena.getOrigin();
        Vector3f delta  = mS.arena.getDelta();
        origin.scale(mS.arenaSize);
        delta.scale(mS.arenaSize);
        
        min_node_width = mS.segmentWidth * 2;
        root = new Node(origin, delta, 0);
        
        insertMesh(mS.arena, mS.arenaSize);
    }
    
    /** Dodaje do drzewa segment */
    public void insertSegment(Segment seg) {
        Vector3f[] vs = seg.vertices;
        root.insertTriangle(
                new Triangle(new Vector3f[] { vs[0], vs[1], vs[2] }, seg));
        root.insertTriangle(
                new Triangle(new Vector3f[] { vs[0], vs[2], vs[3] }, seg));
    }
    
    /** Dodaje do drzewa model */
    public void insertMesh(MeshData mesh, float scale) {
        for (MeshData.Buffers b : mesh.bufs) {
            final int numFaces = b.vs.length/9;
            for (int faceNum=0; faceNum<numFaces; faceNum++) {
                Vector3f[] triVerts = new Vector3f[3];
                for (int vNum = 0; vNum < 3; ++vNum) {
                    final int off = 9*faceNum + 3*vNum;
                    triVerts[vNum] = 
                            new Vector3f(b.vs[off], b.vs[off+1], b.vs[off+2]);
                    triVerts[vNum].scale(scale);
                }
                root.insertTriangle(new Triangle(triVerts, null));
            }
        }
    }
    
    /** Mówi, czy segment koliduje z czymś w drzewie.
     * @param seg segment, który sprawdzamy
     * @param seg2 segment który ignorujemy przy sprawdzeniu (lub null) 
     * */
    public boolean checkSegment(Segment seg, Segment seg2) {
        if (!root.intersectsVertices(seg.vertices))
            return true;
        return root.checkSegment(seg, seg2);
    }
    
    private final class Triangle {
        public final Vector3f[] vs;
        public final Segment seg;
        
        public Triangle(Vector3f[] vs, Segment seg) {
            this.vs = vs;
            this.seg = seg;
        }
    }
    
    /** 
     * Węzeł - reprezentuje sześcian origin .. origin+delta. 
     */
    private final class Node {
        private Vector3f origin;
        private Vector3f delta;
        private int level;
        
        /* Węzeł może być liściem (wtedy ma tylko segments) lub nieliściem
         * (wtedy ma tylko children). 
         */
        private Node[] children;
        private ArrayList<Triangle> triangles;
        
        public Node(Vector3f origin,
                Vector3f delta,
                int level)
        {
            this.origin = origin;
            this.delta = delta;
            this.triangles = new ArrayList<Triangle>();
            this.level = level;
        }
        
        private boolean isLeaf() {
            return children == null;
        }
        
        private void divide() {
            assert isLeaf();
            
            children = new Node[8];
            Vector3f dHalf = new Vector3f(delta);
            dHalf.scale(0.5f);
            for (int i = 0; i < 2; ++i)
                for (int j = 0; j < 2; ++j)
                    for (int k = 0; k < 2; ++k) {
                        Vector3f newOrigin = new Vector3f(
                                origin.x + i*dHalf.x,
                                origin.y + j*dHalf.y,
                                origin.z + k*dHalf.z);
                        children[2*2*i+2*j+k] = new Node(newOrigin, dHalf, level+1);
                    }
            
            for (Triangle tr: triangles)
                insertTriangle(tr);
        }
        
        public boolean checkSegment(Segment seg, Segment seg2) {
            if (isLeaf()) {
                for (Triangle tr: triangles) {
                    if (seg2 != null && seg2.equals(tr.seg))
                        continue;
                    for (int i = 0; i < 4; ++i) {
                        if (IntersectionDetector.intersectTriangleLine(
                                tr.vs[0], tr.vs[1], tr.vs[2],
                                seg.vertices[i], seg.vertices[(i+1)%4]))
                            return true;                        
                    }
                    for (int i = 0; i < 3; ++i) {
                        if (IntersectionDetector.intersectTriangleLine(
                                seg.vertices[0], seg.vertices[1], seg.vertices[2],
                                tr.vs[i], tr.vs[(i+1)%3]))
                            return true;
                        if (IntersectionDetector.intersectTriangleLine(
                                seg.vertices[0], seg.vertices[2], seg.vertices[3],
                                tr.vs[i], tr.vs[(i+1)%3]))
                            return true;
                    }
                }
                return false;
            } else {
                for (Node ch: children)
                    if (ch.intersectsVertices(seg.vertices) &&
                            ch.checkSegment(seg, seg2))
                        return true;
                return false;
            }
        }
        
        /** Dodaje segment do węzła. */
        public void insertTriangle(Triangle tr) {
            if (isLeaf()) {
                triangles.add(tr);
                if (delta.x > min_node_width &&
                        triangles.size() > MAX_TRIANGLES_PER_NODE+level)
                    divide();
            } else {
                for (Node ch: children)
                    if(ch.intersectsVertices(tr.vs))
                        ch.insertTriangle(tr);
            }
        }
        
        /** 
         * Sprawdza, czy dany trójkąt należy do węzła. W pewnych rzadkich 
         * przypadkach może dawać fałszywe pozytywy.
         */
        private boolean intersectsVertices(Vector3f[] vs) {
            /* Test jest bardzo prosty - patrzymy, czy po którejś współrzędnej
             * wszystkie wierzchołki segmentu są na lewo albo na prawo od nas. 
             */
            boolean bx0 = true, by0 = true, bz0 = true;
            boolean bx1 = true, by1 = true, bz1 = true;
            
            for (Vector3f v: vs) {
                bx0 &= (v.x < origin.x);
                by0 &= (v.y < origin.y);
                bz0 &= (v.z < origin.z);

                bx1 &= (v.x > origin.x + delta.x);
                by1 &= (v.y > origin.y + delta.y);
                bz1 &= (v.z > origin.z + delta.z);
            }
            return !(bx0 || by0 || bz0 || bx1 || by1 || bz1);
        }
    }
}
