/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.data;

import javax.vecmath.Vector3f;
import javax.vecmath.Quat4f;

import org.codehaus.jackson.annotate.JsonIgnore;

/**
 * Pojemnik na dane segmentu.
 * 
 * Zakładamy, że dane składowe się nie zmieniają (trudno to zapewnić na 
 * poziomie języka i trzeba na to uważać)
 */
public final class Segment {
    /** Położenie początku segmentu. */
    public final Vector3f pos;
    
    /** 
     * Orientacja na początku segmentu. 
     * Wektory: (0,1,0) - "góra", (0,0,1) - "przód" są obracane o ten 
     * kwaternion.
     */
    public final Quat4f heading;   
        
    /**
     * Kierunek skrętu w danym segmencie. Zachodzi 
     *   seg_n.heading * seg_n.direction.getQuat(...) = seg_(n+1).heading
     */
    public final Direction direction;
    
    /** Czy segment został obliczony przez serwer. */
    public final boolean valid;
    
    /** Tablica dwóch wektorów normalnych */
    @JsonIgnore
    public Vector3f[] normals;
    
    /**
     * Tablica czterech punktów segmentu:
     * <pre>
     *
     * 
     *          /\
     *         /||\      
     *          ||
     *    v[3]-------v[2]
     *      |      // |
     *      |    //   |
     *      |  //     |
     *      |//       |
     *    v[0]---*---v[1] 
     *          pos
     * </pre>
     * 
     * Na razie v[2] i v[3] NIE odpowiadają v[0] i v[1] następnego segmentu.
     */
    @JsonIgnore
    public Vector3f[] vertices;

    /** Wektor przesunięcia segmentu */
    @JsonIgnore
    public Vector3f delta;
    
    /** Orientacja na końcu segmentu. */
    @JsonIgnore
    public Quat4f nextHeading;
    
    private int hash;
    
    public Segment(Vector3f pos, Quat4f heading, Direction direction,
            boolean valid, MatchSettings mS) {
        this.pos = pos;
        this.heading = heading;
        this.direction = direction;
        this.valid = valid;
        calcData(mS);        
        
        this.hash = Util3d.hashVec(pos);
    }
    
    private void calcData(MatchSettings mS) {
        delta = new Vector3f(0, 0, mS.segmentLength);
        Util3d.applyQuatToVector(delta, heading);
        
        nextHeading = new Quat4f(heading);
        Quat4f dirQuat = direction.getQuat(mS.angleX, mS.angleY, mS.angleZ);
        nextHeading.mul(dirQuat);
        
        vertices = new Vector3f[] {
            new Vector3f(-mS.segmentWidth/2, 0, 0),
            new Vector3f( mS.segmentWidth/2, 0, 0),
            new Vector3f( mS.segmentWidth/2, 0, 0),
            new Vector3f(-mS.segmentWidth/2, 0, 0),
        };
        for (int i = 0; i < 4; ++i) {
            Util3d.applyQuatToVector(vertices[i], 
                    i < 2 ? heading : nextHeading);
            vertices[i].add(pos);
            if (i >= 2)
                vertices[i].add(delta);
        }
        
        normals = new Vector3f[] {
            new Vector3f(0, -1, 0),
            new Vector3f(0, -1, 0),
        };
        Util3d.applyQuatToVector(normals[0], heading);
        Util3d.applyQuatToVector(normals[1], nextHeading);
    }
    
    private boolean collidesWithLine(Vector3f v1, Vector3f v2) {
        Vector3f[] vs = this.vertices;
        return IntersectionDetector.intersectTriangleLine(
                    vs[0], vs[1], vs[2], v1, v2) ||
               IntersectionDetector.intersectTriangleLine(
                    vs[0], vs[2], vs[3], v1, v2);
    }
    
    public boolean collidesWith(Segment segment) {
        boolean collides = false;
        
        for (int i = 0; i < 4; ++i) {
            collides |= segment.collidesWithLine(
                    vertices[i], vertices[(i+1)%4]);
            collides |= this.collidesWithLine(
                    segment.vertices[i], segment.vertices[(i+1)%4]);
        }
        return collides;
    }
    
    public Segment calcNext(Direction dir, boolean valid, MatchSettings mS) {
        Vector3f nextPos = new Vector3f(pos);
        nextPos.add(delta);
        return new Segment(
                nextPos, 
                nextHeading, 
                dir, 
                valid,
                mS);
    }

    /** Policz wierzchołki kawałka segmentu o długości tFrac całości. */
    public Vector3f[] calcPartVertices(float tFrac) {
        Vector3f[] vs = new Vector3f[] {
            vertices[0],
            vertices[1],
            new Vector3f(vertices[2]),
            new Vector3f(vertices[3])
        };
        
        vs[2].interpolate(vs[1], 1-tFrac);
        vs[3].interpolate(vs[0], 1-tFrac);
        
        return vs;
    }
    
    /** Policz wektory normalne kawałka segmentu o długości tFrac całości. */
    public Vector3f[] calcPartNormals(float tFrac) {
        Vector3f[] ns = new Vector3f[] {
            normals[0],
            new Vector3f(normals[1])
        };
        ns[1].interpolate(ns[0], 1-tFrac);
        return ns;
    }
    
    public int hashCode() {
        return hash;
    }

    public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (obj.getClass() != Segment.class)
            return false;
        Segment seg = (Segment) obj;
        if (seg.hash != this.hash)
            return false;
        
        return seg.pos.equals(this.pos) && seg.heading.equals(this.heading) &&
            seg.delta.equals(this.delta) && seg.nextHeading.equals(this.nextHeading);
    }
    
    public String toString() {
        return String.format("Segment<%d, %s>", hash, direction);
    }
}
