/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.data.arena;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashSet;

import javax.vecmath.*;

import com.sun.org.apache.regexp.internal.recompile;
import com.sun.xml.internal.ws.api.pipe.NextAction;

import netacka3d.data.IntersectionDetector;
import netacka3d.data.Util3d;
import netacka3d.misc.Logger;

public class Mesh {

    private ArrayList<Point3f> points;
    private ArrayList<Vector3f> normals;
    private ArrayList<Vector2f> texCoords;
    
    private HashSet<Point3f> startingPointsSet;
    public ArrayList<Point3f> startingPointsList;
    
    public float[] bounds;
    
    public ArrayList<Material> mats;
    private int currentMatId = 0;
    private int matCount = 1;
    
    public ArrayList<ArrayList<Tri>> trisByMat;
    
    public Mesh() {
        points = new ArrayList<Point3f>();
        normals = new ArrayList<Vector3f>();
        texCoords = new ArrayList<Vector2f>();
        
        startingPointsSet = new HashSet<Point3f>();
        startingPointsList = new ArrayList<Point3f>();
        
        bounds = new float[6];
        
        trisByMat = new ArrayList<ArrayList<Tri>>();
        trisByMat.add(new ArrayList<Tri>());
        
        mats = new ArrayList<Material>();
        mats.add(new Material("default"));
    }
    
    /* Trójkąt */
    public class Tri {
        public Vertex[] vertices;
        public int matId;
        public Tri(Vertex a, Vertex b, Vertex c) {
            vertices = new Vertex[3];
            vertices[0] = a;
            vertices[1] = b;
            vertices[2] = c;
            
            // ustawianie / sprawdzanie normali
            Vector3f triNormal = calcNormal();
            for (Vertex v: vertices) {
                if (v.normal == null)
                    v.normal = triNormal;
                else if (!v.normal.epsilonEquals(triNormal, 1))
                    // porównanie z bardzo dużym epsilonem, żeby się nie pluło
                    // dla gładkich ścian - ważne, żeby orientacja była ok
                    Logger.w("dziwny normal, oczekiwałem " + triNormal +
                             ", jest " + v.normal);
            }
        }
        
        private Vector3f calcNormal() {
            Vector3f normal = Util3d.crossDelta(
                    vertices[1].point, vertices[0].point,
                    vertices[2].point, vertices[0].point);
            normal.normalize();
            return normal;
        }
    }
    
    /** Przechowuje dane punktu, tj. położenie, wsp. tekstury i normal */
    public class Vertex {
        public Point3f point;
        public Vector2f texCoord;
        public Vector3f normal;
        
        public Vertex(int vIndex, int vtcIndex, int vnIndex) {
            assert vIndex > 0;
            assert vnIndex >= 0;
            assert vtcIndex >= 0;
            point = Mesh.this.points.get(vIndex - 1);
            if (vtcIndex > 0)
                texCoord = Mesh.this.texCoords.get(vtcIndex - 1);
            if (vnIndex > 0) 
                normal = Mesh.this.normals.get(vnIndex - 1);
        }
    }
    
    /** Dzieli ścianę na trójkąty i dodaje ją do modelu. */
    public class FaceBuilder {
        private Vertex first;
        private Vertex last;
        private int i = 0;
                
        public void addVertex(int vIndex, int vtcIndex, int vnIndex) {
            Vertex v = new Vertex(vIndex, vtcIndex, vnIndex);
            if (i==0) first = v;
            else if (i==1) last = v;
            else {
                // uwaga: obiekty first i last mogą być użyte kilka razy 
                // w różnych trójkątach!
                trisByMat.get(currentMatId).add(new Tri(first, last, v));
                last = v;
            }
            i++;
        }
        
        public void commit() {
            // jeśli mamy >= 3 vertexy, to znaczy, że już był wrzucony trójkąt
            // (co najmniej 1), więc commit nie musi wtedy nic robić
            if (i==2) {
                if (startingPointsSet.add(first.point))
                    startingPointsList.add(first.point);
                if (startingPointsSet.add(last.point))
                    startingPointsList.add(last.point);
            }
        }
    }
    
    public static class Material implements Serializable {
        private static final long serialVersionUID = -5538140520670006994L;
        public int id = -1;
        public String name;
        public Tuple3f ambientColor = new Vector3f(0.5f, 0.5f, 0.5f);
        public Tuple3f diffuseColor = new Vector3f(1.0f, 1.0f, 1.0f);
        public Tuple3f specularColor;
        public String ambientTexture;
        public String diffuseTexture;
        public String specularTexture;
        public float transparency;
        public float specComponent; // specular component of the Phong shading
                                    // model, ranges between 0 and 1000
        public int illuminationModel; // 0: no lighting
                                      // 1: diffuse lighting only
                                      // 2: diffuse + specular
        public Material(String name) {
            assert name != null;
            this.name = name;
        }
    }
    
    // ---------------------------------------------------------------
    
    public void addPoint(Point3f v) {
        if (points.isEmpty()) {
            bounds[0] = v.x;
            bounds[1] = v.y;
            bounds[2] = v.z;
            bounds[3] = v.x;
            bounds[4] = v.y;
            bounds[5] = v.z;
        }
        else {
            if (v.x < bounds[0]) bounds[0] = v.x;
            if (v.y < bounds[1]) bounds[1] = v.y;
            if (v.z < bounds[2]) bounds[2] = v.z;
            if (v.x > bounds[3]) bounds[3] = v.x;
            if (v.y > bounds[4]) bounds[4] = v.y;
            if (v.z > bounds[5]) bounds[5] = v.z;
        }
        points.add(v);
    }
    
    public void addNormal(Vector3f v) {
        normals.add(v);
    }
    
    public void addTexCoord(Vector2f v) {
        texCoords.add(v);
    }
    
    public void setCurrentMaterial(Material m) {
        if (m.id == -1) {
            m.id = matCount;
            matCount++;
            mats.add(m);
            trisByMat.add(new ArrayList<Tri>());
            assert(mats.size() == matCount);
        }
        currentMatId = m.id;
        assert(mats.get(currentMatId).id == currentMatId);
    }
    
    // ---------------------------------------------------------------
    
    public MeshData getData() {
        
        MeshData meshData = new MeshData();
        
        meshData.bufs = new MeshData.Buffers[mats.size()];
        for (int matId=0; matId<mats.size(); ++matId) {
            ArrayList<Tri> tris = trisByMat.get(matId);
            
            MeshData.Buffers b = new MeshData.Buffers();
            b.vs = new float[tris.size()*3*3]; // punkty=3*trójkąty, floaty=3*punkty
            b.ns = new float[tris.size()*3*3]; // j.w.
            b.tcs = new float[tris.size()*2*3]; // tu potrzebujemy 2 na punkt
            b.material = mats.get(matId);
            
            int faceNum = 0;
            for (Tri tri : tris) {
                for (int vNum = 0; vNum < 3; ++vNum) {
                    Vertex v = tri.vertices[vNum];
                    b.vs[9*faceNum+3*vNum]   = v.point.x;
                    b.vs[9*faceNum+3*vNum+1] = v.point.y;
                    b.vs[9*faceNum+3*vNum+2] = v.point.z;
                    b.ns[9*faceNum+3*vNum]   = v.normal.x;
                    b.ns[9*faceNum+3*vNum+1] = v.normal.y;
                    b.ns[9*faceNum+3*vNum+2] = v.normal.z;
                    if (v.texCoord != null) {
                        b.tcs[6*faceNum+2*vNum]   = v.texCoord.x;
                        b.tcs[6*faceNum+2*vNum+1] = v.texCoord.y;
                    } else {
                        b.tcs[6*faceNum+2*vNum]   = 0;
                        b.tcs[6*faceNum+2*vNum+1] = 0;
                    }
                }
                faceNum++;
            }
            
            meshData.bufs[matId] = b;
        }
        
        meshData.startingPoints = new float[startingPointsList.size()*3];
        for (int i=0; i<startingPointsList.size(); i++) {
            Point3f p = startingPointsList.get(i);
            meshData.startingPoints[3*i]   = p.x; 
            meshData.startingPoints[3*i+1] = p.y; 
            meshData.startingPoints[3*i+2] = p.z; 
        }
        
        assert bounds.length == 6;
        meshData.bounds = bounds;
        
        return meshData;
    }
}
