/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.logic;

import java.io.IOException;

import netacka3d.data.ArenaData;
import netacka3d.data.Direction;
import netacka3d.data.MatchSettings;
import netacka3d.data.Player;
import netacka3d.data.PlayerColor;
import netacka3d.data.PlayersTable;
import netacka3d.data.Segment;
import netacka3d.logic.Game.MatchState;
import netacka3d.misc.Logger;
import netacka3d.network.ClientNetworkListener;
import netacka3d.network.ClientNetworkNode;

public class ClientController extends Controller {

    protected Game game;
    
    private String nick;
    private PlayerColor preferredColor;
    
    private NetListener netListener;
    private ClientNetworkNode net;
    
    private Direction myDirection = Direction.STRAIGHT;
    
    protected boolean collisionsNeeded = false;
    
    
    
    private LagCompensation lagCompensation;
    
    private String addr;
    private int port;
    
    private GameThread gameThread;

    private boolean broken = false;
    
    /**
     * Odpala watek gry z podanymi ustawieniami (i on sie laczy z serwerem 
     * i robi wszystko inne). 
     * Tę metodę wołamy tylko raz. 
     */
    public void connect(String nick, PlayerColor preferredColor, 
                        String addr, int port) {

        Logger.i("zawołano connect(" + nick + ", " + preferredColor + ", ...)");

        assert !broken;
        assert game == null; // kontroler jest jednorazowy!
        game = new Game();

        assert nick != null && !nick.equals("");
        assert preferredColor != null;
        assert addr != null;
        this.nick = nick;
        this.preferredColor = preferredColor;
        this.addr = addr; 
        this.port = port;
        
        lagCompensation = new LagCompensation();
        
        gameThread = new GameThread();
        gameThread.setName("C_" + this.nick);
        gameThread.start();
    }
    
    /** Kaze watkowi gry sie rozlaczyc i zakonczyc wszystko. */
    public void disconnect() {
        Logger.i("zawołano disconnect(). Czekanie na zakończenie wątku klienta...");
        assert !broken;
        //assert game != null; // wołaliśmy wcześniej connect
        //assert !hasFinished();
        gameThread.state = ClientController.State.FINISHED;
    }
    
    /**
     * Zmiana kierunku przez gracza.
     * @param dir nowy kierunek wstegi gracza
     */
    public void setPlayerDirection(Direction dir) {
        assert(dir != null);
        myDirection = dir;
    }
    
    /* ---------------------------------------------------------------- */
    
    private static enum State { RUNNING, FINISHED, NETWORK_FAILURE }
    
    protected class GameThread extends Thread {
        
        private static final int WORK_TIMEOUT = 1000;
        
        public volatile ClientController.State state = 
                    ClientController.State.RUNNING;

        @Override
        public void run() {
            try {
                go();
            } catch (AssertionError e) {
                broken = true;
                throw e;
            } catch (Exception e) {
                broken = true;
                throw new RuntimeException(e);
            }
        }
            
        private void go() {
            
            assert game != null;
            
            // stawianie sieci
            try {
                netListener = new NetListener();
                net = new ClientNetworkNode(netListener);
                if (!net.start(addr, port))
                    // nic nie robimy, bo onNetFailure bylo juz zawolane
                    return;
            } catch (IOException e) {
                Logger.logException(e);
                notifyStop(e.toString());
                return;
            }

            // dolaczanie "naszego" gracza do gry
            net.join(nick, preferredColor);
            
            // main loop!
            while (state == ClientController.State.RUNNING) {
                int firstMissingSeg = lagCompensation.onStep();
                if (firstMissingSeg > 0) {
                    Logger.i("proszę o resend segmentów od "+firstMissingSeg+
                             " (game.stepNum="+game.stepNum+")");
                    net.resend(firstMissingSeg);
                }

                int timeToWait = lagCompensation.getTimeout();
                net.work(game, Math.min(WORK_TIMEOUT, timeToWait));
                //net.work(game, WORK_TIMEOUT);
            }

            if (state == ClientController.State.FINISHED) {
                notifyStop("finished");
                net.shutdown();
            }
        }
        
    }

    /* ---------------------------------------------------------------- */
    
    /** W tej klasie obslugujemy komunikaty przychodzace z sieci.
     * To wszystko bedzie wolane z watku gry, z net.work(). */
    private class NetListener implements ClientNetworkListener {

        @Override
        public void onJoined(int id) {
            // ignorujemy powtórne przyjście pakietu
            // (protokół nie przewiduje żeby ID się zmieniało)
            if (myPlayerId != 0)
                return;
            myPlayerId = id;
            Logger.i("join się udał, dostałem id " + id);
            lagCompensation = new LagCompensation();
            Logger.d("Wolam notifyStart()...");
            notifyStart();
            Logger.d("koniec notifyStart()...");
        }

        @Override
        public void onJoinFailed() {
            assert(myPlayerId == 0);
            notifyStop("join failed");
        }

        @Override
        public void onNetFailure(String msg) {
            notifyStop(msg);
            gameThread.state = ClientController.State.NETWORK_FAILURE;
        }

        @Override
        public void onNewGameState(int t, MatchSettings matchSettings,
                                   PlayersTable pT, MatchState mState) {
            assert(matchSettings != null);
            assert(pT != null);
            assert(mState != null);
            
            // jeśli nie dostaliśmy jeszcze JOINED, upominamy się o niego 
            // ponownie
            if (myPlayerId == 0)
                net.join(nick, preferredColor);
            
            game.stepNum = t;
            
            Logger.i("onNewGameState (start): nowy stan: "+mState);
            
            try {
                matchSettings.updateArenaFrom(game.matchSettings);
            } catch (IOException e) {
                onNetFailure("can't load arena");
                return;
            }
            
            if (mState == MatchState.ROUND_WARMUP ||
                    game.matchState == MatchState.ROUND_WARMUP) {
                lagCompensation = new LagCompensation(matchSettings.stepRate);
            }
            
            game.matchState = mState;
            game.matchSettings = matchSettings;
            game.players = pT;
            
            /*
            // TO NIE DZIAŁA - DLACZEGO?
            assert pT.getPlayer(myPlayerId) != null; // to powinno być ok, a nie zawsze jest!
            if (!pT.getPlayer(myPlayerId).nickname.equals(nick)) {
                nick = pT.getPlayer(myPlayerId).nickname;
                Logger.i("mam nowy nick: " + nick);
                gameThread.setName("C_" + nick);
            }
            */
            
            if (game.arenaData == null || mState == MatchState.ROUND_WARMUP) {
                game.arenaData = new ArenaData(matchSettings, collisionsNeeded);
                for (Player p : pT.getActivePlayers())
                    game.arenaData.addPlayer(p.id);
            }
            
            nextStep(true);
            
            Logger.d("koniec onNewGameState");
        }

        @Override
        public void onNewSegments(int t, Segment[] segs) {
            Logger.d("dostalem nowe segmenty z sieci, t="+t+
                     ", game.stepNum="+game.stepNum);

            // było: assert segs.length == game.arenaData.ribbonCount();
            if (segs.length != game.arenaData.ribbonCount()) {
                Logger.w("tablica segmentów ma złą długość, olewam ten pakiet");
                return;
            }
            
            lagCompensation.onReceivedSegment(t);
            
            if (t > game.stepNum) {
                game.stepNum = t;
                nextStep(false);
            }
            
            int i = 0;
            for (Player p : game.players.getActivePlayers()) {
                assert game.arenaData.hasPlayer(p.id);
                game.arenaData.setSegment(p.id, t, segs[i]); // wrzucanie
                if (segs[i] != null) // jak ktoś zginął, to tu jest null
                    notifySegment(p.id, t, segs[i]); // powiadamianie
                i++;
            }
        }
        
        /** powiadamia listenerów i wysyła kierunek do sieci */
        private void nextStep(boolean stateChanged) {
            // to wszystko sprawia, ze admin jest o 1 krok do przodu
            notifyStep(stateChanged, game);
            
            // może być tak, że dostajemy stan jeszcze zanim dostaniemy JOINED
            if (myPlayerId != 0)
                net.key(myDirection);
        }
        
    }
}
