/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.logic;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import netacka3d.data.ArenaData;
import netacka3d.data.Direction;
import netacka3d.data.MatchSettings;
import netacka3d.data.Player;
import netacka3d.data.RoundTime;
import netacka3d.data.Segment;
import netacka3d.data.Player.State;
import netacka3d.misc.Logger;

/**
 * Klasa odpowiedzialna za serwerową część logiki gry.
 * 
 * Tylko ta klasa:
 * - zmienia stany graczy
 * - zmienia stany gry
 * 
 * Nie ma wyróżnionego "podłączonego" gracza.
 */
public class ServerGame extends Game {
    
    // odziedziczone z Game:
    // matchState, roundNum, stepNum, players, matchSettings, arenaData
    
    /** nowe (obowiązujące od następnej rundy) ustawienia meczu */
    private MatchSettings newMatchSettings;
    
    // długości poszczególnych etapów meczu (w krokach)
    private int matchWarmupSteps;
    private int roundWarmupSteps;
    private int roundFinishSteps;
    private int matchFinishSteps;
    
    private int numAliveConnectedPlayersInThisRound;
    
    private boolean playersTableChangedInLastStep = false;
    
    public ServerGame(MatchSettings matchSettings) {
        super();
        this.matchSettings = matchSettings;
        try {
            matchSettings.loadArena();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        
        startMatchState(MatchState.MATCH_WARMUP);
        
        float sr = matchSettings.stepRate;
        matchWarmupSteps = (int)(matchSettings.matchWarmupTime * sr) + 1;
        roundWarmupSteps = (int)(matchSettings.roundWarmupTime * sr) + 1;
        roundFinishSteps = (int)(matchSettings.roundFinishTime * sr) + 1;
        matchFinishSteps = (int)(matchSettings.matchFinishTime * sr) + 1;
    }
    
    
    /* ==================== Obsługa kolejnych kroków gry ========== */
    
    /** 
     * Dla danych kierunków graczy liczy kolejny krok.
     * W stanie ROUND podajemy kierunki, w innych nie!
     * @return czy stan się zmienił
     */
    public boolean doStep(Map<Integer, Direction> dirs) {
        assert (matchState == MatchState.ROUND);
        ++stepNum;
        Logger.d("doStep, ROUND, step "+stepNum);
        boolean playersTableChanged = doSegments(dirs);
        boolean matchStateChanged = checkMatchState();
        return playersTableChanged || matchStateChanged;
    }
    public boolean doStep() {
        assert (matchState != MatchState.ROUND);
        ++stepNum;
        return checkMatchState();
    }
    
    /**
     * Policz jeden krok nowych segmentów, uwzględniając kolizje.
     * @return czy stan się zmienił (tylko PlayersTable może się tu zmienić)
     */
    private boolean doSegments(Map<Integer, Direction> dirs) {
        
        assert(stepNum >= 1);
        assert(matchState == MatchState.ROUND);
        
        boolean playersTableChanged = false;
        TreeMap<Integer, Segment> segs = new TreeMap<Integer, Segment>(); 
        
        for (Player segPlayer : players.getActivePlayers()) {
            
            // sprawdzamy tylko dla żywych i podłączonych graczy
            if (!segPlayer.isPlaying() || !segPlayer.isAlive(stepNum))
                continue;
            
            // policz nowy segment
            Direction dir = dirs.get(segPlayer.id);
            if (dir == null) // jeśli gracz nie podał kierunku... 
            	dir = Direction.STRAIGHT; // ...to leci dalej prosto
            
            Segment seg = arenaData.calcNewSegment(segPlayer.id, stepNum,
                                                   dir, true);
            segs.put(segPlayer.id, seg);
            
            // sprawdzanie kolizji
            float coll = arenaData.checkCollision(seg, segPlayer.id);
            if (coll >= 0) {
            	segPlayer.death = new RoundTime(stepNum, coll);
            	numAliveConnectedPlayersInThisRound--;
            	Logger.i("t="+stepNum+", "+segPlayer+" zginął");
            	for (int id : arenaData.getPlayerIds())
            	    if (players.getPlayer(id).isAlive(stepNum+1))
            	        players.getPlayer(id).score++;  
            	playersTableChanged = true;
            }
            
            // wstawianie segmentu do ArenaData
            arenaData.setSegment(segPlayer.id, stepNum, seg);
        }
        
        return playersTableChanged;
    }
    
    
    /* ==================== Automat  ===================== */
    
    /** 
     * Sprawdź odpowiednie warunki i wykonaj ewentualne zmiany stanu meczu.
     * (opis automatu w docu)
     * @return czy stan się zmienił
     */
    private boolean checkMatchState() {
        MatchState oldMatchState = matchState;
        
        switch (matchState) {
        
        case MATCH_WARMUP:
            /* XXX Na razie dopuszczamy 1 gracza w grze. */
            if (stepNum >= matchWarmupSteps && players.getPlayerCount() >= 1)
                startMatchState(MatchState.ROUND_WARMUP);
            break;
            
        case ROUND_WARMUP:
            if (stepNum >= roundWarmupSteps)
                startMatchState(MatchState.ROUND);
            break;
            
        case ROUND:
            // wychodzimy z round, jeśli:
            // - jest mniej niż 2 żywych, a gra więcej niż 1 
            // - jedyny gracz zginął
            int count = arenaData.ribbonCount(),
                alive = numAliveConnectedPlayersInThisRound;
            assert count == players.getActivePlayers().size();
            if ((count > 1 && alive < 2) || (count <= 1 && alive == 0))
                startMatchState(MatchState.ROUND_FINISHED);
            break;
            
        case ROUND_FINISHED:
            if (stepNum >= roundFinishSteps) {
                players.removeDisconnected();
                if (players.getMaxScore() >= matchSettings.scoreLimit ||
                        players.getActivePlayers().size() == 0)
                    startMatchState(MatchState.MATCH_FINISHED);
                else
                    startMatchState(MatchState.ROUND_WARMUP);
            }
            break;
            
        case MATCH_FINISHED:
            if (stepNum >= matchFinishSteps)
                startMatchState(MatchState.MATCH_WARMUP);
            break;
        }
        
        boolean stateChanged = (matchState != oldMatchState);
        if (playersTableChangedInLastStep) {
            playersTableChangedInLastStep = false;
            stateChanged = true;
        }
        return stateChanged;
    }
    
    /** Zainicjuj nowy stan meczu. */
    private void startMatchState(MatchState matchState) {

        Logger.i("t=" + stepNum + ", " + this.matchState + " -> " + matchState);
        
        // uwaga: to jest jedyne miejsce, gdzie this.matchState się zmienia! 
    	this.matchState = matchState;
        stepNum = 0;
        
        switch (matchState) {
        
        case MATCH_WARMUP: // nowy mecz
            roundNum = 0;
            for (Player player: players.getPlayers())
                player.score = 0; // zerowanie punktacji
            break;
            
        case ROUND_WARMUP: // nowa runda
            ++roundNum;
            
            // aktualizacja MatchSettings
            if (newMatchSettings != null) {
                matchSettings = newMatchSettings;
                newMatchSettings = null;
            }
            
            // tworzymy nowe ArenaData
            arenaData = new ArenaData(matchSettings, true);
            
            // aktualizacja PlayersTable
            players.removeDisconnected();
            players.resetDeaths();
            
            // wybieramy aktywnych graczy na czas rundy
            numAliveConnectedPlayersInThisRound = 0;
            for (Player p : players.getPlayers()) {
                p.active = (p.state == Player.State.PLAYING);
                if (p.active) {                 
                    arenaData.addPlayer(p.id);
                    numAliveConnectedPlayersInThisRound++;
                }
            }
            
            arenaData.putFirstSegments();
            break;
            
        case ROUND_FINISHED:
            for (Player p : players.getActivePlayers()) {
                // sprawdzenie, czy nic się nie zmieniło od początku rundy
                assert arenaData.hasPlayer(p.id); 
            }
            assert numAliveConnectedPlayersInThisRound >= 0;
            break;

        default:
            break;
        }

        players.checkConsistency();
    }
    
    
    /* ==================== Inne operacje ==================== */
    
    /** 
     * Ustaw nowe ustawienia meczu.
     * Zmiana wejdzie w życie z nową rundą.
     */
    public void changeMatchSettings(MatchSettings newMatchSettings) {
        this.newMatchSettings = newMatchSettings;
        try {
            this.newMatchSettings.updateArenaFrom(this.matchSettings);
        } catch (IOException e) {
            // TODO porządna obsługa błędu 
            throw new RuntimeException(e);
        }
    }
    
    /** 
     * Dodaj nowego gracza.
     * Zostanie dołączony jako nieaktywny i zacznie grać w następnej rundzie.
     * Zakładamy, że gracza o takim id jeszcze nie ma. 
     */
    public void addPlayer(Player player) {
        players.addPlayer(player);
        playersTableChangedInLastStep = true;
    }
    
    /**
     * Usun gracza z gry (np. kiedy się rozłączył)
     */
    public void disconnectPlayer(int playerId) {
        Player p = players.getPlayer(playerId);
        p.state = State.DISCONNECTED;
        if (matchState == MatchState.ROUND && arenaData.hasPlayer(p.id))
            numAliveConnectedPlayersInThisRound--;
        if (p.death == null)
            p.death = new RoundTime(stepNum, 0);
        // gracz zostanie usunięty w ROUND_FINISHED
    }

}
