/*
 * This file is a part of Netacka 3D.
 *
 * Copyright 2010
 * Leszek Bartkowski, Mikołaj Dądela, Paweł Marczewski, Jan Szejko
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


package netacka3d.network;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.LinkedList;
import java.util.List;

import javax.vecmath.Quat4f;
import javax.vecmath.Vector3f;

import netacka3d.data.Direction;
import netacka3d.data.MatchSettings;
import netacka3d.data.Segment;
import netacka3d.data.serialization.Serialization;
import netacka3d.misc.Logger;


/**
 * Abstrakcyjna klasa wspomagająca moduł sieciowy w kwestii implementacji
 * protokołu. Udostępnia statyczne metody do tworzenia komunikatów do wysłania
 * i identyfikowania odebranych
 * @author Leszek
 *
 */
abstract class UDPProtocol {

    //    static final byte GET_INFO = 1;
    //    static final byte GAME_INFO = 2;
    //    static final byte JOIN = 3;
    //    static final byte NOT_JOINED = 4;
    //    static final byte NEW_GAME_STATE = 5;
    static final byte SEGMENTS = 1;
    static final byte KEY = 2;
    static final byte RESEND = 3;
    static final byte CONFIRM_STATE = 4;
    static final byte PART = 5;

    static final byte NEW_GAME_STATE = 8;
    static final byte GAME_INFO = 9;
    static final byte JOIN = 10;
    static final byte NOT_JOINED = 11;
    static final byte JOINED = 12;
    static final byte BYE = 13;
    static final byte GET_GAME_INFO = 14;
    
    static final byte JSON = 7;

    static final byte NO_DIR = 27;
    //    static final byte DIRS = 4;


    static final int DEFAULT_PORT = 4567;

    static final int SERVER_TRIES = 10;
    static final long SERVER_TIMEOUT = 1000;
    static final long CLIENT_TIMEOUT = 10000;



    /**
     * Tworzy komunikat z segmentami i dopisuje go do bufora
     * @param buf bufor
     * @param t numer kroku gry, którego dotyczą segmenty
     * @param seg segmenty do wysłania
     */
    static void newSegments(ByteBuffer buf,
            int t, Segment[] seg) {
        buf.put(SEGMENTS);
        buf.putInt(t);
        int size = seg.length;
        Segment s;
        for (int i=0; i<size; ++i) {
            s = seg[i];
            if (s==null)
                buf.put(NO_DIR);
            else
                putSegment(buf, s);
        }
    }

    /**
     * Zczytuje segmenty.
     * Uwaga: zakładamy, że z bufora czas t został już wyciągnięty
     * @param buf bufor z segmentami
     * @param settings ustawienia meczu
     * @return
     */
    static Segment[] getJustSegments(ByteBuffer buf, MatchSettings settings) {
        List<Segment> segs = new LinkedList<Segment>();
        while (buf.hasRemaining())
            segs.add(getSegment(buf, settings));
        return segs.toArray(new Segment[0]);
    }



    private static void putVector3f(ByteBuffer buf, Vector3f vec) {
        float x = vec.x, y=vec.y, z=vec.z;
        buf.putFloat(x);
        buf.putFloat(y);
        buf.putFloat(z);
    }

    private static Vector3f getVector3f(ByteBuffer buf) {
        float x,y,z;
        x = buf.getFloat();
        y = buf.getFloat();
        z = buf.getFloat();
        return new Vector3f(x, y, z);
    }

    private static void putQuat4f(ByteBuffer buf, Quat4f qua) {
        buf.putFloat(qua.x);
        buf.putFloat(qua.y);
        buf.putFloat(qua.z);
        buf.putFloat(qua.w);
    }

    private static Quat4f getQuat4f(ByteBuffer buf) {
        float x,y,z,w;
        x = buf.getFloat();
        y = buf.getFloat();
        z = buf.getFloat();
        w = buf.getFloat();        
        return new Quat4f(x,y,z,w);
    }

    /**
     * wkłada do bufora niepusty segment
     * @param buf
     * @param seg
     */
    private static void putSegment(ByteBuffer buf, Segment seg) {
        Direction dir = seg.direction;
        Vector3f pos = seg.pos;
        Quat4f heading = seg.heading;
        buf.put((byte)dir.getIndex());  
        putVector3f(buf, pos);
        putQuat4f(buf, heading);    
        //        Logger.i("PUTSEGMENT: " + seg);
    }

    /**
     * Zczytuje segment z bufora. Zwraca null, jeśli odczytany kierunek
     * będzie NO_DIR (oznaczający brak segmentu)
     * @param buf
     * @param settings
     * @return null jeśli odczytano NO_DIR, wpp Segment
     */
    private static Segment getSegment(ByteBuffer buf, MatchSettings settings) {
        int dirIndex = buf.get();
        if (dirIndex == NO_DIR)
            return null;
        Direction dir = Direction.fromIndex(dirIndex);
        Vector3f pos = getVector3f(buf);
        Quat4f heading = getQuat4f(buf);
        Segment ret =  new Segment(pos, heading, dir, true, settings);
        //        Logger.i("GETSEGMENT: " + ret);
        return ret;
    }

    public static <T> T getJson(ByteBuffer buf, 
            Class<T> type) throws IOException  {
        short length = buf.getShort();
        int pos = buf.position();
        buf.position(pos + length);
//        System.err.println(String.format("Odbior dl: %d, pos: %d", length, buf.position()));
        return Serialization.mapper.readValue(buf.array(), 
                pos, length, type);
    }

    public static <T> void putJson(ByteBuffer buf, T obj) {
        try {
            byte bytes[] = Serialization.mapper.writeValueAsBytes(obj);
            short length = (short) bytes.length;
//            System.err.println(String.format("Dl: %d, dl: %d", bytes.length,length));
            buf.putShort(length);
            buf.put(bytes);

        } catch (IOException e) {
            Logger.logException(e);
            throw new RuntimeException(e);
        }

    }


    //    /**
    //     * normalizuje double do inta, by przeslac go potem przez siec
    //     * @param x
    //     * @return int
    //     */
    //    private static int normalize(float x) {
    ////        int(f/FLOAT_LIMIT*INT32_MAX)
    //        x *= Integer.MAX_VALUE;
    //        return (int) (x/Float.MAX_VALUE*Integer.MAX_VALUE);
    //    }
    //
    //    /**
    //     * otrzymanego z sieci inta denormalizuje uzyskując doublea
    //     * @param x
    //     * @return double
    //     */
    //    private static float denormalize(int x) {
    //        return x*Float.MAX_VALUE/Integer.MAX_VALUE;
    //    }
}